#include <stdio.h>
#include <string>
#include <time.h>
#include <dirent.h>
#include <sys/stat.h>
using namespace std;
#pragma hdrstop  // precompiled headers stop here


// Mail History -- (c) 2002 Lucian Wischik. Everyone can do whatever they
// like with this software, apart from claim ownership. www.wischik.com/lu
// This is version 1.01, incorporating fixes from Ross Younger
// (previously it failed to print December!) Thanks, Ross.
//
// This program reads Berkely Mail Format mailfolders, and prints out statistics
// on the messages inside them. The statistics are cumulative totals for the
// number of messages sent or received, and their volume. (Berkely Mail Format
// is a commonly used format on Unix. It is used for instance by Pine). Note:
// if any individual day sees more than 2GB of email, then this program will
// not count it fully.
//
// I had expected to see a kind of Moore's Law for email (i.e. number of
// messages doubling every year, volume doubling every six months) but it
// didn't seem to work like that for my archive.
//





// These compile-time constants control the size of buffers in the code.
const unsigned int bufsize=16384; // We read the BMF file in this many bytes at a time
const unsigned int linesize=4096; // We recognise From... and Date:... header fields up to this long. NB. Simply a warning, not a crash, if the buffer overflows.

// These things are set up at the start of main()
typedef struct {unsigned int num,size;} TCountInfo;
typedef struct {TCountInfo sent,rcvd;} TDayInfo;
unsigned int CurrentYear;   // e.g. 1997. for Y2K compatability we interpret <1000 years in a range -50 to +950 years centered around current date
int firstyear,numyears;     // the range of years we'll cover
TDayInfo *count;            // count[year-firstyear*12*31 + month*31 + day], where month:0->11 and day:0->31
bool aborted;               // If set, then all slow loops will abort immediately. Not used. I've put it here in case anyone wants to make a GUI version with a Cancel button.
bool terse;                 // For whether to suppress informational messages


// These functions make up the program proper:
int main(int argc, char* argv[]);   // Parses command-line, then reads files, then prints answers
void usage();                       // Prints the command-line options for this program
void ScanFile(const string s);      // Given the name of a file or directory, this either ReadFiles or ScanWilds it.
void ScanWild(const string path,const string wild); // Given e.g. ~/docs/*.bmf, this calls ScanFile on every matching file or directory
void ReadFile(const string fn);     // Given the name of a file, this attempts to process it as BMF
void msg(bool rcvd, const string from,const string date,unsigned int len); // Adds this message into the global count[] array

// General utility functions:
bool ExtractDate(const string date, struct tm *time,bool strict);
string ExtractFilePath(const string FileName);
string ExtractFileName(const string FileName);
bool wildmatch(const char *fn,const char *wild,int reserved); // Does the filename match the wildcard? (reserved must be 0)
bool checkfrom(const char *line);       // Is this a valid From .... line (used as separator in BMF files)?




//---------------------------------------------------------------------------

// main() -- the entry point of this program. Parses the command line,
// then processes all the specified files, then prints the results.
//
int main(int argc, char* argv[])
{ aborted=false; terse=false;
  if (argc<=1) {usage(); return 0;}
  int cmd=1;
  bool countsent=true, countrcvd=true;
  firstyear=1970; numyears=201;
  while (cmd<argc)
  { if (argv[cmd][0]!='-') break;
    if (strncmp(argv[cmd],"-years",6)==0)
    { int lastyear; int i = sscanf(argv[cmd],"-years=%i-%i",&firstyear,&lastyear);
      if (i!=2) {usage(); return 0;}
      numyears = lastyear+1-firstyear;
    }
    else if (strncmp(argv[cmd],"-all",4)==0) {countsent=true;countrcvd=true;}
    else if (strncmp(argv[cmd],"-sent",5)==0) {countsent=true;countrcvd=false;}
    else if (strncmp(argv[cmd],"-rcvd",5)==0) {countsent=false;countrcvd=true;}
    else if (strncmp(argv[cmd],"-terse",6)==0) {terse=true;}
    else if (strncmp(argv[cmd],"-verbose",8)==0) {terse=false;}
    else {usage(); return 0;}
    cmd++;
  }
  if (cmd==argc) {usage(); return 0;}
  //
  time_t tsecs = time(NULL);
  struct tm *t = localtime(&tsecs);
  CurrentYear = 1900 + t->tm_year;
  //
  if (aborted) return 0;
  if (!terse)
  { printf("Doing years %i-%i\n",firstyear,firstyear+numyears-1);
    if (countsent&&countrcvd) printf("Counting all messages, both sent and received\n");
    else if (countsent && !countrcvd) printf("Counting sent messages\n");
    else if (!countsent && countrcvd) printf("Counting received messages\n");
    printf("Any years <1000 will be assumed offset from %i\n",CurrentYear-50);
  }
  count = new TDayInfo[numyears*12*31];
  memset(count,0,sizeof(TDayInfo)*numyears*12*31);

  while (!aborted && cmd<argc)
  { string s = argv[cmd];
    string path=ExtractFilePath(s), name=ExtractFileName(s);
    if (path=="") path=".";
    if (name=="") name="*";
    if (name.find("*")!=string::npos) ScanWild(path,name);
    else ScanFile(s);
    cmd++;
  }

  unsigned int cnum=0, chi=0, clo=0;
  if (!aborted && !terse)
  { printf("---\n");
    printf("day\tmonth\tyear\tc.num\tc.size\n");
  }
  for (int y=0; !aborted && y<numyears; y++)
  { unsigned int year = y+firstyear;
    for (unsigned int m=0; !aborted && m<12; m++)
    { unsigned int month=m+1;
      for (unsigned int d=0; !aborted && d<31; d++)
      { unsigned int day=d+1;
        TDayInfo &inf = count[y*12*31 + m*31 + d]; bool got=false;
        unsigned int oclo=clo;
        if (countsent && inf.sent.num>0) {cnum+=inf.sent.num; clo+=inf.sent.size; got=true;}
        if (countrcvd && inf.rcvd.num>0) {cnum+=inf.rcvd.num; clo+=inf.rcvd.size; got=true;}
        if (got)
        { if (clo<oclo) chi++; // an overflow must have occurred.
          // NB. we assume there's not more than 2GB of email a day.
          if (chi==0) printf("%02i\t%02i\t%04i\t%u\t%u\n",day,month,year,cnum,clo);
          else printf("%02i\t%02i\t%04i\t%u\t%u%u\n",day,month,year,cnum,chi,clo);
        }
      }
    }
  }
  if (!aborted && !terse) printf("\n");
  delete[] count;
  return 0;
}


// usage() -- prints out instructions on how to use this program!
void usage()
{ printf("Usage: mailhistory [-years=1970-2170] [-sent|-rcvd] [-terse] files...\n"
         "* Shows the cumulative number and sizes of your emails.\n"
         "* Understands Berkely Mail Format files. If you specify a directory\n"
         "  name, all files in that directory will be scanned. It will\n"
         "  silently skip over any file that's not in BMF.\n"
         "* By default it will add sent and received together, but you\n"
         "  can limit it to only sent or only received.\n"
         "* By default it will display progress messages; you can suppress\n"
         "  these with the -terse option.\n");
}



//---------------------------------------------------------------------------


// ScanFile - we're given a filename for either a directory or a file.
// If it's a directory, we scan all the files inside it. If it's a file,
// we process it directly.
void ScanFile(const string s)
{ if (aborted) return;
  struct stat fs;
  int res = stat(s.c_str(),&fs);
  if (res!=0)
  { if (terse) fprintf(stderr,"Unable to read file %s\n",s.c_str());
    else printf("Unable to read file %s\n",s.c_str());
    return;
  }
  if (fs.st_mode&S_IFDIR) ScanWild(s,"*");
  else ReadFile(s);
}

// ScanWild - we're given a directory, and a wildcard to match within this directory
// For every file we find that matches the wildcard, we ScanFile it.
void ScanWild(const string path,const string wild)
{ if (aborted) return;
  if (!terse)
  { if (wild=="*") printf("%s/\n",path.c_str());
    else printf("%s/%s\n",path.c_str(),wild.c_str());
  }
  DIR *dir = opendir(path.c_str());
  if (dir==NULL)
  { if (terse) fprintf(stderr,"Unable to read directory %s\n",path.c_str());
    else printf("Unable to read directory %s\n",path.c_str());
    return;
  }
  struct dirent *ent = readdir(dir);
  while (!aborted && ent!=NULL)
  { string fn = ent->d_name;
    if (fn!="." && fn!="..")
    { if (wildmatch(fn.c_str(),wild.c_str(),0)) ScanFile(path+"/"+fn);
    }
    ent = readdir(dir);
  }
  closedir(dir);
}

// ReadFile -- given the filename of an actual file, we attempt to process
// it as a BMF file. If no message can be recognised from it at all, then
// we return silently. Otherwise, as soon as we've recognised our first
// message, then we print the filename.
void ReadFile(const string fn)
{ bool isbmf=false; // Have we recognised at least one valid message?
  FILE *f = fopen(fn.c_str(),"r");
  // First, a shortcut. The first letter in the BMF file must be 'F',
  // as part of the line "From ...". If we can't find this, then we
  // can abort immediately, no need for anything more complicated.
  int cc = '\n'; while (cc=='\n' || cc=='\r' || cc==' ' || cc=='\t') cc=fgetc(f);
  if (cc!='F') {fclose(f);return;} // not a valid BMF file

  // buf[] -- we read the file in this many bytes at a time. Therefore
  // our inner loop can just retrieve the next character from this
  // buffer, instead of having to invoke the library function fgetc every
  // single time.
  char buf[bufsize];
  // line[] -- when we think we're recognising the From or Date: line, then
  // we steadly accumulate it inside this buffer. Note: if the header line
  // should exceed the size of the line[] buffer, it won't cause a crash:
  // just a warning.
  char line[linesize];
  // from, date, len -- these contain the attributes of the message we're
  // currently parsing. (i.e. it's From line, its Date: line, and its length).
  string from="", date=""; unsigned int len=0;


  // We will use a finite-state automaton to recognise messages
  // inside a Berkely Mail Format file.
  // A message starts with [\r]\n[\r]\nFrom person@somewhere 15:23:27[\r]\n
  // and might have one or more headings of the form [\r]\nReceived: ...[\r]\n
  // and might have a date of the form [\r]\nDate: ...[\r]\n
  // Then it has a header-separator [\r]\n[\r]\n
  // and finishes with the next message or the end of the file
  //
  // state 0:  nothing                     "\n"->1, else->0
  // state 1:  \n                          "\r"->2, "\n"->3, else->0
  // state 2:  \n \r                       "\n"->3, else->0
  // state 3:  \n [\r] \n                  "F"->4, "\n"->3, "\r"->2, else->0
  // state 4:  \n [\r] \n F                "r"->5, else->0
  // state 5:  \n [\r] \n Fr               "o"->6, else->0
  // state 6:  \n [\r] \n Fro              "m"->7, else->0
  // state 7:  \n [\r] \n From             " "->FSTART,8, else->0
  // state 8:  \n [\r] \n From_*           "\r"->9, "\n"->FEND,10, else->8
  // state 9:  \n [\r] \n From_* \r        "\n"->FEND,10, else->8
  // state 10: hdrs \n                     "\n"->3, "\r"->11, "R"->14, "D"->21, else->12
  // state 11: hdrs \n \r                  "\n"->3, else->12
  // state 12: hdrs \n *                   "\n"->10, "\r"->13, else->12
  // state 13: hdrs \n * \r                "\n"->10, else->12
  // state 14: hdrs \n R                   "e"->15, "\r"->13, "\n"->10, else->12
  // state 15: hdrs \n Re                  "c"->16, "\r"->13, "\n"->10, else->12
  // state 16: hdrs \n Rec                 "e"->17, "\r"->13, "\n"->10, else->12
  // state 17: hdrs \n Rece                "i"->18, "\r"->13, "\n"->10, else->12
  // state 18: hdrs \n Recei               "v"->19, "\r"->13, "\n"->10, else->12
  // state 19: hdrs \n Receiv              "e"->20, "\r"->13, "\n"->10, else->12
  // state 20: hdrs \n Receive             "d"->RCVD,12, "\r"->13, "\n"->10, else->12
  // state 21: hdrs \n D                   "a"->22, "\r"->13, "\n"->10, else->12
  // state 22: hdrs \n Da                  "t"->23, "\r"->13, "\n"->10, else->12
  // state 23: hdrs \n Dat                 "e"->24, "\r"->13, "\n"->10, else->12
  // state 24: hdrs \n Date                ":"->DSTART,26, "\r"->13, "\n"->10, else->12
  // state 25: *removed*
  // state 26: hdrs \n Date:*              "\r"->27, "\n"->DEND,10, else->26
  // state 27: hdrs \n Date:* \r           "\n"->DEND,10, else->26

  unsigned int state=4; // initial state is "recognised the first F of the file"
  int lpos=-1; // This is either -1, or an index into the line[] array for where to stash the next character we read
  bool fstart=false,fend=false,rcvd=false,dstart=false,dend=false,bstart=false; // A bunch of
  // variables that are set by the finite-state-machine, to tell the next bit of
  // code whether it need take any special action.

  while (!aborted)
  { size_t red = fread(buf,1,bufsize,f);
    const char *c=buf, *max=buf+red;
    while (c<max && !aborted)
    { char d = *c;
      switch (state)
      { case 0: if(d=='\n')state=1; break;
        case 1: if(d=='\r')state=2;else if(d=='\n')state=3;else state=0; break;
        case 2: if(d=='\n')state=3;else state=0; break;
        case 3: if(d=='F')state=4;else if(d=='\n')state=3;else if(d=='\r')state=2;else state=0; break;
        case 4: if(d=='r')state=5;else state=0; break;
        case 5: if(d=='o')state=6;else state=0; break;
        case 6: if(d=='m')state=7;else state=0; break;
        case 7: if(d==' '){fstart=true;state=8;}else state=0; break;
        case 8: if(d=='\r')state=0;else if(d=='\n'){fend=true;state=10;}else state=8; break;
        case 9: if(d=='\n'){fend=true;state=10;}else state=8; break;
        case 10:if(d=='\n'){bstart=true;state=3;}else if(d=='\r')state=11;else if(d=='R')state=14;else if(d=='D')state=21;else state=12; break;
        case 11:if(d=='\n'){bstart=true;state=3;}else state=12; break;
        case 12:if(d=='\n')state=10;else if(d=='\r')state=13;else state=12; break;
        case 13:if(d=='\n')state=10;else state=12; break;
        case 14:if(d=='e')state=15;else if(d=='\r')state=13;else if(d=='\n')state=10;else state=12; break;
        case 15:if(d=='c')state=16;else if(d=='\r')state=13;else if(d=='\n')state=10;else state=12; break;
        case 16:if(d=='e')state=17;else if(d=='\r')state=13;else if(d=='\n')state=10;else state=12; break;
        case 17:if(d=='i')state=18;else if(d=='\r')state=13;else if(d=='\n')state=10;else state=12; break;
        case 18:if(d=='v')state=19;else if(d=='\r')state=13;else if(d=='\n')state=10;else state=12; break;
        case 19:if(d=='e')state=20;else if(d=='\r')state=13;else if(d=='\n')state=10;else state=12; break;
        case 20:if(d=='d'){rcvd=true;state=12;}else if(d=='\r')state=13;else if(d=='\n')state=10;else state=12; break;
        case 21:if(d=='a'||d=='A')state=22;else if(d=='\r')state=13;else if(d=='\n')state=10;else state=12; break;
        case 22:if(d=='t'||d=='T')state=23;else if(d=='\r')state=13;else if(d=='\n')state=10;else state=12; break;
        case 23:if(d=='e'||d=='E')state=24;else if(d=='\r')state=13;else if(d=='\n')state=10;else state=12; break;
        case 24:if(d==':'){dstart=true;state=26;}else if(d=='\r')state=13;else if(d=='\n')state=10;else state=12; break;
        case 26:if(d=='\r')state=27;else if(d=='\n'){dend=true;state=10;}else state=26; break;
        case 27:if(d=='\n'){dend=true;state=10;}else state=26;break;
        default:
        { state=0; if (!isbmf) {fclose(f);return;} else
          { if (terse) fprintf(stderr,"Error recognising file... may not be BMF\n");
            else printf("Error recognising file... may not be BMF\n");
          }
        }
      }
      c++;
      //
      if (fstart) {lpos=0;fstart=false;}
      else if (fend)
      { line[lpos]=0;lpos=-1;fend=false;
        if (checkfrom(line))
        { if (from!="") {msg(rcvd,from,date,len); if (!isbmf && !terse) printf("%s\n",fn.c_str()); isbmf=true;}
          const char *c=line; while (*c==' ' || *c=='\t') c++; from=c;
          rcvd=false; date=""; len=0;
        }
      }
      else if (dstart) {lpos=0; dstart=false;}
      else if (dend) {line[lpos]=0; const char *c=line; if (*c==':') c++; while (*c==' ' || *c=='\t') c++; date=c; lpos=-1; dend=false;}
      else if (bstart) {len=0;bstart=false;}
      else len++;
      if (lpos!=-1)
      { line[lpos]=d;lpos++;
        if (lpos>=(int)linesize)
        { line[lpos-1]=0; lpos=-1;
          if (terse) fprintf(stderr,"Header too long: %s\n",line);
          else printf("Header too long: %s\n",line);
        }
      }
    }
    if (red==0) break;
  }
  msg(rcvd,from,date,len);

  fclose(f);
}


// msg -- when we've just parsed an entire message, we'll add it into the count,
// after checking that its headers are okay
void msg(bool rcvd, const string from,const string date,unsigned int len)
{ if (from=="" || from=="MAILER-DAEMON" || aborted) return;
  struct tm t;
  if (date=="")
  { if (terse) fprintf(stderr,"Missing Date: ignoring message (%s)\n",from.c_str());
    else printf("Missing Date: ignoring message (%s)\n",from.c_str());
    return;
  }
  bool okaydate = ExtractDate(date,&t,false);
  if (!okaydate)
  { if (terse) fprintf(stderr,"Malformed Date: \"%s\"... ignoring message (%s)\n",date.c_str(),from.c_str());
    else printf("Malformed Date: \"%s\"... ignoring message (%s)\n",date.c_str(),from.c_str());
    return;
  }
  if (t.tm_year<firstyear || t.tm_year>=firstyear+numyears)
  { if (terse) fprintf(stderr,"Year %i out of range... ignoring message (%s)\n",t.tm_year,from.c_str());
    else printf("Year %i out of range... ignoring message (%s)\n",t.tm_year,from.c_str());
    return;
  }
  TDayInfo &day = count[(t.tm_year-firstyear)*12*31 + t.tm_mon*31 + t.tm_mday-1];
  TCountInfo *c; if (rcvd) c=&day.rcvd; else c=&day.sent;
  c->num++; c->size+=len;
}

//---------------------------------------------------------------------------



int lu_strnicmp(const char *a,const char *b,int n)
{ while (true)
  { if (n==0) return 0;
    if (*a==0 && *b==0) return 0;
    if (*a==0) return -1;
    if (*b==0) return 1;
    char ca=*a, cb=*b;
    if (ca>='A' && ca<='Z') ca|=32;
    if (cb>='A' && cb<='Z') cb|=32;
    if (ca<cb) return -1;
    if (ca>cb) return 1;
    a++; b++; n--;
  }
}


// ExtractDate -- Given a date in any of a variety of formats, we return it as a struct tm.
// We set the fields year, mon, mday, hour, min, sec as follows:
// year: the actual year (e.g. 2003), not year-1900 as is used by the RTL functions.
// month: 0->11
// mday: 0->30
// hour: 0->23
// min: 0->59
// sec: 0->59
// and return true/false for whether it was succesful.
// NB. If time is absent, then we return hour=0/min=0/sec=0 and true.
bool ExtractDate(const string date, struct tm *time,bool strict)
{ // Example formats:
  // [Wed,] 14 Jul 1999 18:23:59 +0100 (GMT)
  // 1/14/00 6:50:06 PM Pacific Daylight Time
  // [Wed] Apr 19 20:33:32 2000
  // 07 Feb 102 18:23:59 +0000
  //
  // Hence: we split it into blocks, and judge what each block might be.
  // If it is a three-block of letters (perhaps trailing comma) then it's
  // either a month-name or a weekname. We can tell which.
  // If it is a block with two colons and otherwise numbers, it is a time.
  // If it is a block with two slashes and otherwise numbers, it is a date.
  // If it is a number, and is the first number we've received, it is a day-of-month
  // If it is a number, and is the second number we've received, it is a year
  // If we have got time and year, then stop.
  //

  int day,month,year,hour,min,sec;
  //
  const char *c=date.c_str(), *max=c+date.length(); while (*c==' ') c++;
  bool gottime=false, gotday=false, gotmonth=false, gotyear=false;
  while (c<max && (!gottime || !gotyear || !gotday || !gotmonth))
  { int colons=0, slashes=0; bool allnums=true; const char *startc=c;
    while (c<max && *c!=',' && *c!=' ' && *c!='\n' && *c!='\r' && *c!='\t')
    { if (*c==':') colons++;
      else if (*c=='/') slashes++;
      else if (*c<'0' || *c>'9') allnums=false;
      c++;
    }
    const char *endc=c; int len=c-startc;
    if (*c==',') c++; while (c<max && (*c==' ' || *c=='\n' || *c=='\r' || *c=='\t')) c++;
    //
    if (!allnums && len>=3)
    { if (len==3 && !allnums) // the month? or could be a weekday...
      { string s(startc,len); for (string::size_type i=0; i<s.length(); i++) s[i]=(char)tolower(s[i]);
        bool thismonth=true;
        if (s=="jan") month=1; else if (s=="feb") month=2; else if (s=="mar") month=3;
        else if (s=="apr") month=4; else if (s=="may") month=5; else if (s=="jun") month=6;
        else if (s=="jul") month=7; else if (s=="aug") month=8; else if (s=="sep") month=9;
        else if (s=="oct") month=10; else if (s=="nov") month=11; else if (s=="dec") month=12;
        else if (s=="gen") month=1; else if (s=="mai") month=5; else if (s=="giu") month=6;
        else if (s=="lug") month=7; else if (s=="set") month=9; else if (s=="ott") month=10;
        else if (s=="dic") month=12;
        else thismonth=false;
        if (thismonth) gotmonth=true;
        if (strict && !thismonth)
        { bool thisday=false;
          if (s=="mon" || s=="tue" || s=="wed" || s=="thu" || s=="fri" || s=="sat" || s=="sun") thisday=true;
          if (s=="lun" || s=="mar" || s=="mer" || s=="gio" || s=="ven" || s=="sab" || s=="dom") thisday=true;
          if (!thisday) return false; // unrecognised chunk
        }
      }
      else if (len==4 && !allnums)
      { if (lu_strnicmp(startc,"july",4)==0) {month=7;gotmonth=true;}
        else if (lu_strnicmp(startc,"june",4)==0) {month=6;gotmonth=true;}
        else if (strict) return false; // unrecognised
      }
      else if (len==5 && !allnums)
      { if (lu_strnicmp(startc,"march",5)==0) {month=3;gotmonth=true;}
        else if (lu_strnicmp(startc,"april",5)==0) {month=4;gotmonth=true;}
        else if (strict) // could be +0100 or (GMT)
        { bool isdif=false, iszone=false;
          if ((startc[0]=='+' || startc[0]=='-')
              && (startc[1]>='0' && startc[1]<='9')
              && (startc[2]>='0' && startc[2]<='9')
              && (startc[3]>='0' && startc[3]<='9')
              && (startc[4]>='0' && startc[4]<='9')) isdif=true;
          if (startc[0]=='(' && startc[4]==')'
              && startc[1]>='A' && startc[1]<='Z'
              && startc[2]>='A' && startc[2]<='Z'
              && startc[3]>='A' && startc[3]<='Z'
              && startc[4]>='A' && startc[4]<='Z') iszone=true;
          if (!isdif && !iszone) return false; // unrecognised
        }
      }
      else if (len==6 && !allnums)
      { if (lu_strnicmp(startc,"august",6)==0) {month=8;gotmonth=true;}
        else if (strict) return false;
      }
      else if (len==7 && !allnums)
      { if (lu_strnicmp(startc,"january",7)==0) {month=1;gotmonth=true;}
        else if (lu_strnicmp(startc,"october",7)==0) {month=10;gotmonth=true;}
        else if (strict) return false;
      }
      else if (len==8 && !allnums)
      { if (lu_strnicmp(startc,"february",8)==0) {month=2;gotmonth=true;}
        else if (lu_strnicmp(startc,"november",8)==0) {month=11;gotmonth=true;}
        else if (lu_strnicmp(startc,"december",8)==0) {month=12;gotmonth=true;}
        else if (strict) return false;
      }
      else if (len==9 && !allnums)
      { if (lu_strnicmp(startc,"september",9)==0) {month=9;gotmonth=true;}
        else if (strict) return false;
      }
    }
    //
    else if (colons>=1 && allnums) // the time
    { const char *d=startc;
      while (*d=='0') d++; const char *startd=d; while (*d!=':') d++; string shour(startd,d-startd); d++;
      while (*d=='0' && d<endc) d++; startd=d; while (d<endc && *d!=':') d++; string smin(startd,d-startd); if (d<endc) d++;
      string ssec="0";
      if (colons==2)
      { while (*d=='0' && d<endc) d++; startd=d; while (d<endc) d++; ssec=string(startd,d-startd);
      }
      if (shour=="") shour="0"; sscanf(shour.c_str(),"%i",&hour);
      if (smin=="") smin="0";   sscanf(smin.c_str(),"%i",&min);
      if (ssec=="") ssec="0";   sscanf(ssec.c_str(),"%i",&sec);
      gottime=true;
    }
    //
    else if (slashes==2 && allnums) // the full date in american format
    { const char *d=startc;
      while (*d=='0') d++; const char *startd=d; while (*d!='/') d++; string smonth(startd,d-startd); d++;
      while (*d=='0') d++; startd=d; while (*d!='/') d++; string sday(startd,d-startd); d++;
      while (*d=='0' && d<endc) d++; startd=d; while (d<endc) d++; string syear(startd,d-startd);
      if (sday=="") sday="0";     sscanf(sday.c_str(),"%i",&day);
      if (smonth=="") smonth="0"; sscanf(smonth.c_str(),"%i",&month);
      if (syear=="") syear="0";   sscanf(syear.c_str(),"%i",&year);
      gotday=true; gotmonth=true; gotyear=true;
    }
    //
    else if (slashes==0 && colons==0 && allnums && !gotday) // the day of the month
    { while (*startc=='0' && startc<endc) startc++;
      string sday(startc,endc-startc); if (sday=="") sday="0";
      sscanf(sday.c_str(),"%i",&day);
      gotday=true;
    }
    //
    else if (slashes==0 && colons==0 && allnums && gotday) // the year
    { while (*startc=='0' && startc<endc) startc++;
      string syear(startc,endc-startc); if (syear=="") syear="0";
      sscanf(syear.c_str(),"%i",&year);
      gotyear=true;
    }
  }

    // have we got any leftover junk at the end of the line?
  if (strict)
  { if (!gottime || !gotyear || !gotday || !gotmonth) return false;
    bool junk=false;
    while (c<max)
    { if (*c!=' ') junk=true;
      c++;
    }
    if (junk) return false;
  }

  bool okay = (gotday && gotmonth && gotyear && 1<=month && month<=12 && 1<=day && day<=31);
  if (!okay) return false;
  //
  if (gotyear && year<1000)
  { int y=(CurrentYear+50) % 100;
    int cen = CurrentYear - (CurrentYear % 100);
    if (year>y) year-=100;
    year += cen;
  }
  if (time!=0)
  { time->tm_year = year;
    time->tm_mon = month-1; // tm_mon is 0->11, but we got month as 1->12
    time->tm_mday = day;    // tm_mday is 1->31
    if (gottime)
    { time->tm_hour = hour;   // tm_hour is 0->23
      time->tm_min = min;
      time->tm_sec = sec;
    }
    else
    { time->tm_hour = 0;
      time->tm_min = 0;
      time->tm_sec = 0;
    }
  }
  return true;
}





// checkfrom -- Berkely Mail Format uses "From Lines" as separators between messages.
// But sometimes a message might contain the word "From" just as text inside it.
// We need to check a bit more carefully whether this line really is a separator,
// and not just body text...
bool checkfrom(const char *lstart)
{ // We expect to find a line "From c.harrington@abdn.ac.uk Wed Nov 15 10:34:05 2000"
  // M. Hardcastle advises that some old mailers use weird ! separators
  // instead of this familiar internet email address.
  // Also, some of them just have "From username <date>" without any at symbol.
  const char *lend=lstart; while (*lend!=0 && *lend!='\r' && *lend!='\n') lend++;
  bool found=true; const char *fc=lstart; while (*fc==' ') fc++;
  while (found && fc<lend && *fc!='@' && *fc!='!' && *fc!=' ') fc++;
  // Now we have got either the @, or a space, after the eaddr/username
  if (*fc!='@' && *fc!='!')
  { // we allow some grace: if there was no @, but a perfect date, we count it.
    // or if it was from the mailer daemon, we count it
    if (strncmp(lstart+5,"MAILER-DAEMON",13)!=0)
    { while (fc<lend && *fc==' ') fc++; if (fc==lend) found=false;
      if (found) found=ExtractDate(string(fc,lend-fc),NULL,true);
    }
  }
  else
  { // by now we have encountered at least one @ separator
    while (found && fc<lend && *fc!=' ') fc++; if (fc==lend) found=false;
    // by now we are at a space after the email address
    while (found && fc<lend && *fc==' ') fc++; if (fc==lend) found=false;
    // by now we are at the start of the date
    while (found && fc<lend && *fc!=':') fc++; if (*fc!=':') found=false;
    while (found && fc<lend && *fc>='0' && *fc<='9') fc++;
    if (*fc!=' ' && *fc!=':') found=false;
    // while normal found means that it was at least something like a date:
    // By this point we've found From xxx@xxx_____:##:___ or :## ____, which is good enough
  }
  return found;
}





int LastDelimiter(const string Delimiters, const string s)
{ for (int i=s.length()-1; i>=0; i--)
  { if (Delimiters.find(s[i])!=string::npos) return i;
  }
  return -1;
}


// ExtractFilePath: extracts the drive and directory parts of the given
// filename. The resulting string is the leftmost characters of FileName,
// up to and including the colon or backslash that separates the path
// information from the name and extension. The resulting string is empty
// if FileName contains no drive and directory parts.
string ExtractFilePath(const string FileName)
{ int i = LastDelimiter("\\/:",FileName);
  if (i==-1) return ""; else return FileName.substr(0,i);
}

// ExtractFileName: extracts the name and extension parts of the given
// filename. The resulting string is the leftmost characters of FileName,
// starting with the first character after the colon or backslash that
// separates the path information from the name and extension. The resulting
// string is equal to FileName if FileName contains no drive and directory parts.
string ExtractFileName(const string FileName)
{ int i=LastDelimiter("\\/:",FileName);
  if (i==-1) return FileName;
  else return FileName.substr(i+1);
}

bool wildmatch(const char *fn,const char *wild,int depth)
{ if (*wild==0 && *fn==0) return true;
  if (*wild==0) return false;
  if (strcmp(wild,"*")==0 && (*fn!='.' || depth>0)) return true;
  if (strcmp(wild,"*.*")==0) {while (*fn!=0 && *fn!='.') fn++; return (*fn=='.');}
  if (*fn==0) return false;
  char w=*wild, f=*fn;
  if (f=='.' && depth==0 && w!='.') return false;
  if (w=='?') return wildmatch(fn+1,wild+1,depth+1);
  if (w=='*') return wildmatch(fn+1,wild,depth+1) | wildmatch(fn,wild+1,depth+1);
  if (w!=f) return false;
  return wildmatch(fn+1,wild+1,depth+1);
}


